/* CDLL: circular doubly-linked list.
   in TBCppA project: tracer-based CPP analyzer.
    Copyright (c) 2006 Katsuhiko Gondow (gondow(at)cs.titech.ac.jp)
    $Id: list.c,v 1.1 2006/06/07 09:29:20 gondow Exp $ */

/* CDLL: circular doubly-linked list.

   The first CDLL element is always used as anchor, i.e., 
   a pointer to the list, and logically not a member of the list.
       null list: CDLL
       singleton: CDLL->CDLL
       3 elems. : CDLL->CDLL->CDLL->CDLL
                 anchor head        tail
    anchor's data is always ignored (not used for any purpose).
    This structure makes this file's code simpler. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "list.h"
#include "misc.h"
/* ================================================================== */
static void CDLL_dump (struct CDLL *head) __attribute__ ((unused));
/* ================================================================== */
static void
CDLL_dump  (struct CDLL *head)
{
    int i;
    struct CDLL *p;
    for (i = 0, p = head->next; p != head; i++, p = p->next) {
	fprintf (stderr, "%d: %s\n", i, p->param_name);
    }
}
/* ================================================================== */
void
CDLL_init   (struct CDLL *head)
{
    head->param_name = NULL;
    head->next = head->prev = head;
}

void
CDLL_insert (struct CDLL *head, char *param_name)
{
    struct CDLL *cdll = malloc (sizeof (struct CDLL));
    assert (cdll != NULL);
    cdll->param_name = param_name;
    cdll->prev = head->prev;
    cdll->next = head;
    head->prev->next = cdll;
    head->prev = cdll;
}

int
CDLL_query  (struct CDLL *head, char *param_name)
{
    int i;
    struct CDLL *p;
    assert (param_name != NULL);
    if (head == NULL)
	return -1;
    for (i = 0, p = head->next; p != head; i++, p = p->next) {
	if (!strcmp (param_name, p->param_name))
	    return i;
    }
    return -1; /* not found */
}

void
CDLL_free   (struct CDLL *head)
{
    struct CDLL *p, *next;
    assert (head != NULL);
    for (p = head->next; p != head; p = next) {
        next = p->next;
	p->next->prev = p->prev;
	p->prev->next = p->next;
	p->next = p->prev = NULL;
	EFREE (p->param_name);
	EFREE (p);
    }
    assert (head->next == head);
    assert (head->prev == head);
}

#if 0
void
CDLL_unparse (struct CDLL *head, FILE *fp)
{
    struct CDLL *p;
    for (p = head->next; p != head; p = p->next) {
	if (p != head->next)
	    fputs (", ", fp);
	fputs (p->param_name, fp);
    }
}
#endif
/* ================================================================== */
#if 0
int
main (void)
{
    struct CDLL head;
    CDLL_init (&head);
    CDLL_insert (&head, estrdup ("param1"));
    CDLL_insert (&head, estrdup ("param2"));
    CDLL_insert (&head, estrdup ("param3"));
    CDLL_insert (&head, estrdup ("param4"));
    CDLL_dump (&head);
    printf ("query: %d\n", CDLL_query (&head, "param1"));
    printf ("query: %d\n", CDLL_query (&head, "param2"));
    printf ("query: %d\n", CDLL_query (&head, "param3"));
    printf ("query: %d\n", CDLL_query (&head, "param4"));
    printf ("query: %d\n", CDLL_query (&head, "param5"));
    CDLL_unparse (&head, stderr);
    CDLL_free (&head);
}
#endif
/* ================================================================== */
