/* misc utility
   in TBCppA project: tracer-based CPP analyzer.
   Copyright (c) 2006   Katsuhiko Gondow (gondow(at)cs.titech.ac.jp)
   $Id: misc.c,v 1.1 2006/06/07 09:29:20 gondow Exp $ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <errno.h>
#include <assert.h>
#include <stdarg.h>

#include "misc.h"
#include "tef.h"
/* ==================================================================== */
/* ==================================================================== */
/* ==================================================================== */
void
chop (char *s)
{
    if (s == NULL) return;
    s [strlen (s) - 1] = '\0';
}

void
chop_nl (char *s)
{
    int len;
    if (s == NULL) return;
    len = strlen (s);
    if (s [len - 1] == '\n')
	s [len - 1] = '\0';
}

int
mkdirhier (char *dir, mode_t mode)
{
    char *slash = dir;
    struct stat sbuf;
    int ret;
    
    if (*slash == '/')
	slash++;

    while (1) {
	slash = strchr (slash, '/');
	if (slash != NULL)
	    *slash = '\0';
	
	ret = stat (dir, &sbuf);
	if (ret == -1) {
	    if (errno != ENOENT) { perror ("stat"); return -1; }
	    ret = mkdir (dir, mode);
	    if (ret == -1 && errno != EEXIST) { perror ("mkdir"); return -1; }
	} else if (!S_ISDIR(sbuf.st_mode)) {
	    fprintf (stderr, "%s: %s\n", dir, strerror (ENOTDIR));
	    return -1;
	}

	if (slash == NULL)
	    break;
	*slash++ = '/';
    }
    return 0;
}

FILE *
fopen_with_mkdir (const char *path, const char *mode)
{
    int ret;
    char *path2 = estrdup (path);
    FILE *fp = fopen (path, mode);

    if (fp == NULL && errno == ENOENT) {
	char *s = strrchr (path2, '/');
	if (s != NULL)
	    *s = '\0';
#if 1
	ret = mkdirhier (path2, 0777);
#else
	{
	    char *cmd;
	    asprintf (&cmd, "mkdir -p %s", path2);
	    ret = system (cmd);
	}
#endif
	if (ret == 0)
	    fp = fopen (path, mode);
	else
	    fp = NULL;
    }

    EFREE (path2);
    return fp;
}

void *
emalloc (int size)
{ 
    void *p = malloc (size);
    if (p == NULL) perror ("malloc");
    return p;
}

void *
erealloc (void *ptr, size_t size)
{
    void *p = realloc (ptr, size);
    if (p == NULL) perror ("realloc");
    return p;
}

void
efree (void *ptr, char *file, int line)
{
    assert (ptr != NULL);
    free (ptr);
}

char *
estrdup (const char *s)
{
#if 0
    char *p = strdup (s);
#else
    int size = strlen (s) + 1;
    char *p = emalloc (size);
    memcpy (p, s, size);
#endif
    if (p == NULL) perror ("strdup");
    return p;
}

#if 0
int
easprintf (char **sp, const char *fmt, ...)
{
    int ret;
    va_list pvar;
    va_start (pvar, fmt);
    ret = vasprintf (sp, fmt, pvar);
    va_end (pvar);
    if (ret == -1) perror ("asprintf");
    return ret;
}
#else
int
easprintf (char **sp, const char *fmt, ...)
{
    int ret, size;
    va_list pvar;
    va_start (pvar, fmt);
    size = vsnprintf (NULL, 0, fmt, pvar);
    va_end (pvar);
    if (size == -1) {
	perror ("vsnprintf"); exit (1);
    }

    va_start (pvar, fmt);
    *sp = emalloc (size + 1);
    ret = vsnprintf (*sp, size + 1, fmt, pvar);
    va_end (pvar);
    if (ret == -1) {
	perror ("vsnprintf"); exit (1);
    }

    return ret;
}
#endif

char *
str_to_xml (char *s)
{
    char *p, *q;
    char *xml;
    int size;

    size = 0;
    for (p = s; *p != '\0'; p++) {
	switch (*p) {
	case '<':  size += strlen ("&lt;");   break;
	case '>':  size += strlen ("&gt;");   break;
	case '&':  size += strlen ("&amp;");  break;
	case '\'': size += strlen ("&apos;"); break;
	case '\"': size += strlen ("&quot;"); break;
	default:   size++;                    break;
	}
    }

    xml = emalloc (size + 1);

    q = xml;
    for (p = s; *p != '\0'; p++) {
	switch (*p) {
	case '<':  strcpy (q, "&lt;");   q += strlen ("&lt;");   break;
	case '>':  strcpy (q, "&gt;");   q += strlen ("&gt;");   break;
	case '&':  strcpy (q, "&amp;");  q += strlen ("&amp;");  break;
	case '\'': strcpy (q, "&apos;"); q += strlen ("&apos;"); break;
	case '\"': strcpy (q, "&quot;"); q += strlen ("&quot;"); break;
	default:   *q = *p;              q++;                    break;
	}
    }
    *q = '\0';
    return xml;
}

void
xml_encode (FILE *fp, char *text)
{
    char *p;

    for (p = text; *p != '\0'; p++) {
	switch (*p) {
	case '<':  fputs ("&lt;", fp);   break;
	case '>':  fputs ("&gt;", fp);   break;
	case '&':  fputs ("&amp;", fp);  break;
	case '\'': fputs ("&apos;", fp); break;
	case '\"': fputs ("&quot;", fp); break;
	default:   fputc (*p, fp);       break;
	}
    }
}

#if 0
/* add the option '-Wl,-wrap=malloc' to gcc */
void *
__wrap_malloc (int size)
{ 
    return __real_malloc (size);
}

void *
__wrap_realloc (void *ptr, size_t size)
{
    return __real_realloc (ptr, size);
}

void
__wrap_free (void *ptr)
{
    __real_free (ptr);
}
#endif

/* ==================================================================== */
