/* tadd: embedding CPP tracers in C programs.
   in TBCppA project: tracer-based CPP analyzer.
   Copyright (c) 2006   Katsuhiko Gondow (gondow(at)cs.titech.ac.jp)
   $Id: tadd.l,v 1.5 2006/06/21 15:59:17 gondow Exp gondow $ */

%{
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "tadd.h"
#include "tadd.tab.h"
#include "misc.h"
#include "list.h"

#define UNUSED __attribute__((unused))

#define YY_USER_ACTION { yylloc.first_line   = yylloc.last_line;        \
                         yylloc.first_column = yylloc.last_column;      \
                         yylloc.last_column += yyleng;                  \
                         yylval.str = yylval.str2 = yylval.tadd = NULL; \
                         yylval.list = NULL; }
#define NEWLINE_UPDATE ( yylloc.last_line++, yylloc.last_column = 0 )
#define STRDUP         ( yylval.str  = estrdup (yytext),  \
                         yylval.str2 = estrdup (yytext),  \
                         yylval.tadd = estrdup (yytext) )

int is_if_or_elif = 0;

static char *string_buf  = NULL;
static int  string_size  = 0;
static char *comment_buf = NULL;
static int  comment_size = 0;

static int comment_first_line, comment_first_column;

static void add_char_to_string  (char c);
static void add_char_to_comment (char c);
%}

%x comment string include define_head
%s if_or_elif

 /* abbreviation definitions ('*' means 'optional') 
      D : decimal digit,   O : octal digit,      H : hexadecimal digit
      L : letter,          E : signed exponent
      IS: integer suffix*, FS: floating suffix*
      WS: whitespace without newline */

O      [0-7]
D      [0-9]
H      [a-fA-F0-9]
IS     (([uU][lL]?)|([uU](ll|LL))|([lL][uU]?)|((ll|LL)[uU]?))
L      [a-zA-Z_$]
FS     [fFlL]?
E      [Ee][+-]?{D}+
WS     [ \t\f\v\r]           

L_INTEGER_CONSTANT  (((0{O}+)|({D}+)|(0[xX]{H}+)){IS}?)
L_FLOATING_CONSTANT (({D}+{E}{FS})|({D}*"."{D}+{E}?{FS})|({D}+"."{D}*{E}?{FS}))

L_SIMPLE_ESCAPE_SEQUENCE \\[ntvbrfa\\\?'"]
L_OCTAL_ESCAPE_SEQUENCE  \\{O}{1,3}
L_HEXADECIMAL_SEQUENCE   \\(x|X){H}+
L_UNIVERSAL_CHARACTER_NAME \\(u{H}{4}|U{H}{8})
L_ESCAPE_SEQUENCE        {L_SIMPLE_ESCAPE_SEQUENCE}|{L_OCTAL_ESCAPE_SEQUENCE}|{L_HEXADECIMAL_SEQUENCE}|{L_UNIVERSAL_CHARACTER_NAME}
L_CHARACTER_CONSTANT     (L?'([^\\\n']|{L_ESCAPE_SEQUENCE})+')
 /*
L_KEYWORD    auto|break|case|char|const|continue|default|do|double|else|enum|extern|float|for|goto|if|int|long|register|return|short|signed|sizeof|static|struct|switch|typedef|union|unsigned|void|volatile|while|inline|restrict|_Bool|_Complex|_Imaginary
 */
L_IDENTIFIER ({L}|{L_UNIVERSAL_CHARACTER_NAME})({L}|{D})*
L_CONSTANT {L_INTEGER_CONSTANT}|{L_FLOATING_CONSTANT}|{L_CHARACTER_CONSTANT}
L_PUNCTUATOR      "["|"]"|"{"|"}"|"."|"->"|"++"|"--"|"&"|"*"|"+"|"-"|"~"|"!"|"/"|"%"|"<<"|">>"|"<"|">"|"<="|">="|"=="|"!="|"^"|"|"|"&&"|"||"|"?"|":"|";"|"..."|"="|"*="|"/="|"%="|"+="|"-="|"<<="|">>="|"&="|"^="|"|="|","|"#"|"##"|"<:"|":>"|"<%"|"%>"|"%:"|"%:%:"

%% /* ==================================================================== */

^{WS}*#{WS}*include  { BEGIN(include); return INCLUDE; }
^{WS}*#{WS}*include_next  { BEGIN(include); return INCLUDE_NEXT; }
<include>{WS}+       ;
<include>\<[^\n>]+\> { STRDUP;         return H_HEADER_NAME; }
<include>\"[^\n"]+\" { STRDUP;         return Q_HEADER_NAME; }
<include>.           { BEGIN(INITIAL); unput(yytext[0]); }
<include>\n    	     { BEGIN(INITIAL); NEWLINE_UPDATE; return NEWLINE;}

^{WS}*#{WS}*define   { BEGIN(define_head); return DEFINE; }
<define_head>{WS}+   ;
<define_head>{L_IDENTIFIER}       {
                       BEGIN(INITIAL);
                       yylval.str  = estrdup (yytext);
                       easprintf (&yylval.str2, "%s%s",
                                  estrdup (yytext), *double_suffix_p);
	               yylval.tadd = estrdup (yylval.str2);
                       return IDENTIFIER; }
<define_head>{L_IDENTIFIER}"("    { 
                     char *str2;
                     BEGIN(INITIAL);
                     yylval.str  = estrdup (yytext);
		     str2 = estrdup (yytext);
		     chop (str2);
                     easprintf (&yylval.str2, "%s%s(",
						str2, *double_suffix_p);
		     EFREE (str2);
	             yylval.tadd = estrdup (yylval.str2);
                     return IDENTIFIER2;
                     }

^{WS}*#{WS}*undef    return UNDEF;
^{WS}*#{WS}*if       { BEGIN(if_or_elif); is_if_or_elif = 1; return IF; }
^{WS}*#{WS}*elif     { BEGIN(if_or_elif); is_if_or_elif = 1; return ELIF;}
<if_or_elif>\n       { BEGIN(INITIAL); is_if_or_elif = 0;
                       NEWLINE_UPDATE; return NEWLINE;}
^{WS}*#{WS}*ifdef    return IFDEF;
^{WS}*#{WS}*ifndef   return IFNDEF;
^{WS}*#{WS}*else     return ELSE;
^{WS}*#{WS}*endif    return ENDIF;

^{WS}*#{WS}*pragma   return PRAGMA;
^{WS}*#{WS}*line     return LINE;
^{WS}*#{WS}*error    return ERROR;
^{WS}*#{WS}*\r?\n    return EMPTY;

"("                  return '(';
")"                  return ')';
","                  return ',';
"..."                return ELLIPSIS;
"#"                  return '#';
"##"                 return SHARPSHARP;
 /* even keyword can be replaced when CPP is working, so comment out is a must
{L_KEYWORD}          { STRDUP; return KEYWORD; }
 */
{L_IDENTIFIER}       { yylval.str  = estrdup (yytext);
                       easprintf (&yylval.str2, "%s%s",
                                  estrdup (yytext), *double_suffix_p);
	               yylval.tadd = estrdup (yylval.str2);
                       return IDENTIFIER; }
{L_CONSTANT}         { STRDUP; return CONSTANT; }

 /* string-literal */
L?\"    	     { BEGIN(string); 
                       string_buf = estrdup ("\"");
		       string_size = 1; }
<string>"\\"[\\\"]   { add_char_to_string (yytext [0]); 
                       add_char_to_string (yytext [1]); }
<string>"\""   	     { BEGIN(INITIAL);
                       add_char_to_string ('\"'); 
                       yylval.str  = estrdup (string_buf);
                       yylval.str2 = estrdup (string_buf);
                       yylval.tadd = estrdup (string_buf);
		       EFREE (string_buf); string_buf = NULL;
		       return STRING_LITERAL; }
<string>"\n"   	     { // admitting illegal string literals
                       add_char_to_string ('\n'); 
                       NEWLINE_UPDATE; }
<string>.    	     { add_char_to_string (yytext [0]); }

{L_PUNCTUATOR}       { STRDUP; return PUNCTUATOR; }
"\\\n"|"\\\r\n"      { NEWLINE_UPDATE; }  /* continuation line */
"\n"                 { NEWLINE_UPDATE; return NEWLINE; }
{WS}+                ;
 
 /* comment */
"/*"	    	     { BEGIN(comment);
                       comment_buf = estrdup ("/*");
                       comment_size = 2; 
                       comment_first_line   = yylloc.first_line;
                       comment_first_column = yylloc.first_column; }
<comment>"*/" 	     { char *xml;
                       BEGIN(INITIAL);
                       add_char_to_comment ('*'); add_char_to_comment ('/');
		       xml = str_to_xml (comment_buf);
                       fprintf (commout, "<comment filename_ref='F%d' "
                                "first_line='%d' first_column='%d' "
                                "last_line='%d'  last_column='%d' "
                                "type='delimited'>%s</comment>\n", idnum,
                                comment_first_line, comment_first_column,
                                yylloc.last_line, yylloc.last_column, xml);
                       EFREE (xml); EFREE (comment_buf); comment_buf = NULL; }
<comment>\n          { add_char_to_comment ('\n'); NEWLINE_UPDATE; }
<comment>.    	     { add_char_to_comment (yytext [0]); }
"//".*               { char *xml = str_to_xml (yytext);
                       fprintf (commout, "<comment filename_ref=\"F%d\" "
                                "first_line=\"%d\" first_column=\"%d\" "
                                "last_line=\"%d\"  last_column=\"%d\" "
                                "type=\"single_line\">%s</comment>\n", idnum,
                                yylloc.first_line, yylloc.first_column,
                                yylloc.last_line,  yylloc.last_column, xml); 
		       EFREE (xml); }
.                    {    STRDUP;
                          if (strchr (yytext, '@') != NULL)
                              fprintf (stderr, "!!!! '@' appeared (%d.%d)\n",
                              yylloc.first_line, yylloc.first_column);
                          return NON_WHITE_SPACE_CHAR;
                     }

%% /* ==================================================================== */
static void
add_char_to_string (char c)
{
    string_buf = erealloc (string_buf, ++string_size + 1);
    string_buf [string_size - 1] = c;
    string_buf [string_size]     = '\0';
}

static void
add_char_to_comment (char c)
{
    comment_buf = erealloc (comment_buf, ++comment_size + 1);
    comment_buf [comment_size - 1] = c;
    comment_buf [comment_size]     = '\0';
}
/* ==================================================================== */
void
tadd_unput (char *s)
{
    char *p = s + strlen (s) - 1;
    while (p >= s) {
	unput (*p--);
    }
}

#if 0
int
main (void)
{
    int token;
    while (token = yylex ()) {
	fprintf (stderr, "token = %d, lexeme = |%s|\n", token, yytext);
    }
}
#endif

