/* tadd: embedding CPP tracers in C programs.
   in TBCppA project: tracer-based CPP analyzer.
   Copyright (c) 2006   Katsuhiko Gondow (gondow(at)cs.titech.ac.jp)
   $Id: tadd.y,v 1.13 2006/06/20 03:54:55 gondow Exp gondow $ */

%{
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <assert.h>
#include <sys/param.h>
#include <signal.h>

#include "tadd.h"
#include "misc.h"
#include "pi.h"
#include "list.h"
#include "tef.h"

#define YYDEBUG 1
#define YYERROR_VERBOSE

#define LINE_FMT      "filename_ref='F%d' first_line='%d' first_column='%d' "\
                      "last_line='%d' last_column='%d' "
#define LINE_INFO(x)  idnum, (x).first_line, (x).first_column, (x).last_line, (x).last_column
#define LINE_INFO2(x,y)  idnum, (x).first_line, (x).first_column, (y).last_line, (y).last_column

#define NULLIFY(y)  ((y).str = (y).str2 = (y).tadd = NULL, (y).list = NULL)
#define MOVE(x,y)   ((x).str  = (y).str,  (x).str2 = (y).str2, \
                     (x).tadd = (y).tadd, (x).list = (y).list, \
                     NULLIFY(y))
#define MOVE1_FMT(x,y,fmt)    (easprintf (&x.str,  fmt, y.str),   \
			       easprintf (&x.str2, fmt, y.str2),  \
			       easprintf (&x.tadd, fmt, y.tadd),  \
			       x.list = NULL, free_lval (&y))
#define MOVE2_FMT(x,y,z,fmt)  (easprintf (&x.str,  fmt, y.str,  z.str),   \
			       easprintf (&x.str2, fmt, y.str2, z.str2),  \
			       easprintf (&x.tadd, fmt, y.tadd, z.tadd),  \
			       x.list = NULL, free_lval (&y), free_lval (&z))
#define BOTTOM(x,y) (x.str  = estrdup (y), x.str2 = estrdup (y), \
                     x.tadd = estrdup (y), x.list = NULL)

FILE *commout;
char *absolute_path;
int  idnum;

static void lex_sync (void);
static void free_lval  (struct lval *lval);
static void free_lval2 (struct lval *lval);
static int  yyerror (char *mesg);
static void print_usage (void);
static void process_options (int argc, char *argv []);
static void output_xml_begin (char *file);
static void output_xml_end   (void);
static void output_params    (struct CDLL *list);
static void output_params2   (struct CDLL *list);
static void tadd (char *file);
static void output_error_text (int idnum, int first_line, int first_column,
			       int last_line, int last_column);

static struct CDLL *cdll_head = NULL;
static int id_counter         = 0;
static int define_func_flag   = 0;
static struct {
    char **file;
    int  file_num;
    char *dir;
    char *db_file;
    char *comm_suffix;
    char *double_suffix;
    int  conservative;
} args = {
    NULL,       /* char **file; */
    0,          /* int  file_num; */
    "./TBCppA", /* char *dir; */
    "db_file",  /* char *db_file; */
    ".comm",    /* char *comm_suffix; */
    "_1528552", /* char *double_suffix; */ /* default is Titech zip code :-) */
    0,          /* int  conservative; */
};

char **double_suffix_p = &args.double_suffix;

%}

%error-verbose
%expect 1
%start group

%token  IDENTIFIER IDENTIFIER2 CONSTANT STRING_LITERAL PUNCTUATOR NEWLINE NON_WHITE_SPACE_CHAR H_HEADER_NAME Q_HEADER_NAME

%token INCLUDE      "#include"
%token INCLUDE_NEXT "#include_next"
%token DEFINE       "#define"
%token UNDEF        "#undef"
%token IF           "#if"
%token IFDEF        "#ifdef"
%token IFNDEF       "#ifndef"
%token ELIF         "#elif"
%token ELSE         "#else"
%token ENDIF        "#endif"
%token ELLIPSIS     "..."
%token SHARPSHARP   "##"
%token LINE         "#line"
%token ERROR        "#error"
%token PRAGMA       "#pragma"
%token EMPTY        "#empty"

%left  SHARPSHARP  /* C99 says the eval order of ## op is unspecified */
%nonassoc '#'

%% /* ==================================================================== */

DF_on:   { lex_sync (); define_func_flag = 1; cdll_head = NULL; } ;
DF_off:  { lex_sync (); define_func_flag = 0; cdll_head = NULL; } ;

group
    : /* empty */      { NULLIFY ($$); }
    | group group_part { NULLIFY ($$); }
    ;

group_part
    : {
	  fprintf (yyout, "@\"cond id='F%d_%d'\"\n", idnum, id_counter);
	  id_counter++;
      } 
      if_group elif_groups else_group "#endif" NEWLINE
      {
	  fprintf (yyout, "#endif\n");
	  fprintf (yyout, "%s%s%s", $2.tadd, $3.tadd, $4.tadd);
	  fprintf (yyout, "@\"/cond\"\n");
	  free_lval (&$2); free_lval (&$3); free_lval (&$4); 
	  NULLIFY($$);
      }
    | "#include" H_HEADER_NAME NEWLINE
      {
	  char *xml = str_to_xml ($2.str);
	  fprintf (yyout, "@\"include id='F%d_%d' type='system' file='%s'\"\n",
		   idnum, id_counter, xml);
	  EFREE (xml);
	  fprintf (yyout, "#include %s\n", $2.str);
	  fprintf (yyout, "@\"/include\"\n");
	  id_counter++; free_lval (&$2);
	  NULLIFY($$);
      }
    | "#include" Q_HEADER_NAME NEWLINE
      {
	  char *xml = str_to_xml ($2.str);
	  fprintf (yyout, "@\"include id='F%d_%d' type='user' file='%s'\"\n",
		   idnum, id_counter, xml);
	  EFREE (xml);
	  fprintf (yyout, "#include %s\n", $2.str);
	  fprintf (yyout, "@\"/include\"\n");
	  id_counter++; free_lval (&$2);
	  NULLIFY($$);
      }
    | "#include" pp_tokens_define  NEWLINE
      {
	  char *xml = str_to_xml ($2.str);
	  fprintf (yyout,
		   "@\"include id='F%d_%d' type='pp_tokens' file='%s'\"\n"
		   "@\"exp orig='%s'\"%s@\"/exp\"\n",
		   idnum, id_counter, xml, $2.str, $2.tadd);
	  EFREE (xml);
	  fprintf (yyout, "#include %s\n", $2.str);
	  fprintf (yyout, "@\"/include\"\n");
	  id_counter++; free_lval (&$2);
	  NULLIFY($$);
      }
    | "#include_next" H_HEADER_NAME NEWLINE
      {
	  char *xml = str_to_xml ($2.str);
	  fprintf (yyout,
		   "@\"include_next id='F%d_%d' type='system' file='%s'\"\n",
		   idnum, id_counter, xml);
	  EFREE (xml);
	  fprintf (yyout, "#include_next %s\n", $2.str);
	  fprintf (yyout, "@\"/include_next\"\n");
	  id_counter++; free_lval (&$2);
	  NULLIFY($$);
      }
    | "#include_next" Q_HEADER_NAME NEWLINE
      {
	  char *xml = str_to_xml ($2.str);
	  fprintf (yyout,
		   "@\"include_next id='F%d_%d' type='user' file='%s'\"\n",
		   idnum, id_counter, xml);
	  EFREE (xml);
	  fprintf (yyout, "#include_next %s\n", $2.str);
	  fprintf (yyout, "@\"/include_next\"\n");
	  id_counter++; free_lval (&$2);
	  NULLIFY($$);
      }
    | "#include_next" pp_tokens_define  NEWLINE
      {
	  char *xml = str_to_xml ($2.str);
	  fprintf (yyout,
		   "@\"include_next id='F%d_%d' type='pp_tokens' file='%s'\"\n"
		   "@\"exp orig='%s'\"%s@\"/exp\"\n",
		   idnum, id_counter, xml, $2.str, $2.tadd);
	  EFREE (xml);
	  fprintf (yyout, "#include_next %s\n", $2.str);
	  fprintf (yyout, "@\"/include_next\"\n");
	  id_counter++; free_lval (&$2);
	  NULLIFY($$);
      }
    | "#define" IDENTIFIER pp_tokens_define_opt NEWLINE
      {
	  char *xml = str_to_xml ($3.str);
	  fprintf (yyout, "@\"define type='obj' id='F%d_%d' "
		   "name='%s' macro_body='%s' " LINE_FMT "/\"\n",
		   idnum, id_counter, $2.str, xml, LINE_INFO(@$));
	  EFREE (xml);
	  fprintf (yyout, "#define %s %s\n", $2.str, $3.str); 

	  fprintf (yyout, "#ifndef %s\n", $2.str2);
	  fprintf (yyout, "#define %s @\"macro_body ref='F%d_%d'\"%s@\"/macro_body\"\n",
		   $2.str2, idnum, id_counter, $3.tadd); 
	  fprintf (yyout, "#endif\n");
	  
	  id_counter++; free_lval (&$2); free_lval (&$3);
	  NULLIFY($$);
      }
    | "#define" IDENTIFIER2 DF_on
      identifier_list_opt ')' pp_tokens_define_opt NEWLINE DF_off 
      { 
	  char *xml = str_to_xml ($6.str);
	  char *id2 = strdup ($2.str);
	  char *double2 = strdup ($2.str2);
	  chop (id2); chop (double2);

	  fprintf (yyout, "@\"define type='func' id='F%d_%d' "
		   "name='%s' args='%s' macro_body='%s' " LINE_FMT "\"\n",
		   idnum, id_counter, id2, $4.str, xml, LINE_INFO(@$));
	  EFREE (xml);
	  EFREE (id2);
	  output_params ($4.list);
	  fprintf (yyout, "@\"/define\"\n");
          fprintf (yyout, "#define %s%s) %s\n", $2.str, $4.str, $6.str);

	  fprintf (yyout, "#ifndef %s\n", double2);
	  EFREE (double2);
	  fprintf (yyout, "#define %s%s%s%s) @\"macro_body ref='F%d_%d'\"",
		   $2.str2, $4.str, ((!strcmp ($4.str, "")) ? "" : ", "),
		   $4.str2, idnum, id_counter);
	  output_params2 ($4.list);
	  fprintf (yyout, "%s@\"/macro_body\"\n", $6.tadd);
	  fprintf (yyout, "#endif\n");

          id_counter++; free_lval (&$2); free_lval2 (&$4); free_lval (&$6);
	  NULLIFY($$);
      }
    | "#define" IDENTIFIER2 DF_on
      "..." ')' pp_tokens_define_opt NEWLINE DF_off
      { 
	  char *xml = str_to_xml ($6.str);
	  char *id2 = strdup ($2.str);
	  char *double2 = strdup ($2.str2);
	  chop (id2); chop (double2);

	  fprintf (yyout, "@\"define type='func_variadic' id='F%d_%d' "
		   "name='%s' args='...' macro_body='%s' " LINE_FMT "/\"\n",
		   idnum, id_counter, id2, xml, LINE_INFO(@$));
	  EFREE (xml);
	  EFREE (id2);
	  fprintf (yyout, "#define %s...) %s\n", $2.str, $6.str);

	  fprintf (yyout, "#ifndef %s\n", double2);
	  EFREE (double2);
	  fprintf (yyout, "#define %s...) @\"macro_body ref='F%d_%d'\""
		   "@\"param name='__VA_ARGS__'\"__VA_ARGS__@\"/param\""
		   "%s@\"/macro_body\"\n",
		   $2.str2, idnum, id_counter, $6.tadd);
	  fprintf (yyout, "#endif\n");

	  id_counter++; free_lval (&$2); free_lval (&$6);
	  NULLIFY($$);
      }
    | "#define" IDENTIFIER2 DF_on
      identifier_list ',' "..." ')' pp_tokens_define_opt NEWLINE DF_off
      {
	  char *xml = str_to_xml ($8.str);
	  char *id2 = strdup ($2.str);
	  char *double2 = strdup ($2.str2);
	  chop (id2); chop (double2);

	  fprintf (yyout, "@\"define type='func_variadic' id='F%d_%d' "
		   "name='%s' args='%s, ...' macro_body='%s' " LINE_FMT "\"\n",
		   idnum, id_counter, id2, $4.str, xml, LINE_INFO(@$));
	  EFREE (xml);
	  EFREE (id2);
	  output_params ($4.list);
	  fprintf (yyout, "@\"/define\"\n");
	  fprintf (yyout, "#define %s%s, ...) %s\n", $2.str,  $4.str, $8.str);

	  fprintf (yyout, "#ifndef %s\n", double2);
	  EFREE (double2);
	  fprintf (yyout,
		   "#define %s%s, %s, ...) @\"macro_body ref='F%d_%d'\"",
		   $2.str2, $4.str, $4.str2, idnum, id_counter);
	  output_params2 ($4.list);
	  fprintf (yyout, 
		   "@\"param name='__VA_ARGS__'\"__VA_ARGS__@\"/param\""
		   "%s@\"/macro_body\"\n", $8.tadd);
	  fprintf (yyout, "#endif\n");

	  id_counter++; free_lval (&$2); free_lval2 (&$4); free_lval (&$8);
	  NULLIFY($$);
      }
    | "#undef"  IDENTIFIER NEWLINE
      {
	  fprintf (yyout, "@\"undef id='F%d_%d' name='%s' " LINE_FMT "/\"\n",
		   idnum, id_counter, $2.str, LINE_INFO(@$));
	  fprintf (yyout, "#undef %s\n", $2.str);
	  fprintf (yyout, "#undef %s\n", $2.str2);
	  id_counter++; free_lval (&$2);
	  NULLIFY($$);
      }
    | "#line"   pp_tokens     NEWLINE
      {
	  fprintf (yyout, "#line   %s\n", $2.str);
      }
    | "#error"  pp_tokens_opt NEWLINE
      {
	  fprintf (yyout, "#error  %s\n", $2.str);
      }
    | "#pragma" pp_tokens_opt NEWLINE
      {
	  fprintf (yyout, "#pragma %s\n", $2.str);
      }
    | "#empty"  NEWLINE
      {
	  fprintf (yyout, "#\n");
      }
    | pp_tokens_opt NEWLINE
      {
	  fprintf (yyout, "%s\n", $1.tadd);
	  free_lval (&$1);
	  NULLIFY($$);
      }
    | error NEWLINE
      {
	  yyerrok;
	  fprintf (yyout, "/* error recovery (F%d:%d.%d-%d.%d) */\n",
		   LINE_INFO(@$));
	  output_error_text (LINE_INFO(@$));
	  fprintf (yyout, "/* end of error recovery */\n");
	  NULLIFY($$);
      }
    ;

if_group
    : "#if"     pp_tokens  NEWLINE 
      { 
	  char *xml = str_to_xml ($2.str);
	  fprintf (yyout, "#if %s\n", $2.str);
          fprintf (yyout, "@\"cond_selected type='if' ref='F%d_%d' "
		   LINE_FMT "\"\n", idnum, id_counter, LINE_INFO2(@1,@3));
	  fprintf (yyout, "@\"exp orig='%s'\"%s@\"/exp\"\n",
		   xml, $2.tadd);
	  EFREE (xml);
	  $$.str = (char *)id_counter++; 
      }
      group
      {
	  char *xml = str_to_xml ($2.str);
	  NULLIFY ($$);
	  easprintf (&$$.tadd,
		     "@\"if id='F%d_%d' exp='%s' " LINE_FMT "/\"\n",
		     idnum, (int)$4.str, xml, LINE_INFO(@$));
	  EFREE (xml);
	  free_lval (&$2);
	  fprintf (yyout, "@\"/cond_selected\"\n");
      }
    | "#ifdef"  IDENTIFIER NEWLINE
      {
	  fprintf (yyout, "#ifdef %s\n",  $2.str);
          fprintf (yyout, "@\"cond_selected type='ifdef' ref='F%d_%d' "
		   LINE_FMT "\"\n", idnum, id_counter, LINE_INFO2(@1,@3));
	  fprintf (yyout, "@\"exp orig='%s'\"%s@\"/exp\"\n",
		   $2.str, $2.tadd);
	  $$.str = (char *)id_counter++; 
      }
      group
      {
	  NULLIFY ($$);
	  easprintf (&$$.tadd,
		     "@\"ifdef id='F%d_%d' exp='%s' " LINE_FMT "/\"\n",
		     idnum, (int)$4.str, $2.str, LINE_INFO(@$));
	  free_lval (&$2);
	  fprintf (yyout, "@\"/cond_selected\"\n");
      }
    | "#ifndef" IDENTIFIER NEWLINE
      {
	  fprintf (yyout, "#ifndef %s\n", $2.str);
          fprintf (yyout, "@\"cond_selected type='ifndef' ref='F%d_%d' "
		   LINE_FMT "\"\n", idnum, id_counter, LINE_INFO2(@1, @3));
	  fprintf (yyout, "@\"exp orig='%s'\"%s@\"/exp\"\n",
		   $2.str, $2.tadd);
	  $$.str = (char *)id_counter++; 
      }
      group
      {
	  NULLIFY ($$);
	  easprintf (&$$.tadd,
		     "@\"ifndef id='F%d_%d' exp='%s' " LINE_FMT "/\"\n",
		     idnum, (int)$4.str, $2.str, LINE_INFO(@$));
	  free_lval (&$2);
	  fprintf (yyout, "@\"/cond_selected\"\n");
      }
    ;

elif_groups
    : /* empty */
      {
	  NULLIFY ($$);
	  $$.tadd = estrdup ("");
      }
    | elif_groups "#elif" pp_tokens NEWLINE
      {
	  char *xml = str_to_xml ($3.str);
	  fprintf (yyout, "#elif %s\n", $3.str);
          fprintf (yyout, "@\"cond_selected type='elif' ref='F%d_%d' "
		   LINE_FMT "\"\n", idnum, id_counter, LINE_INFO2(@2,@4));
	  fprintf (yyout, "@\"exp orig='%s'\"%s@\"/exp\"\n",
		   xml, $3.tadd);
	  EFREE (xml);
	  $$.str = (char *)id_counter++; 
      }
      group
      {
	  char *xml = str_to_xml ($3.str);
	  NULLIFY ($$);
	  easprintf (&$$.tadd,
		     "%s@\"elif id='F%d_%d' exp='%s' " LINE_FMT "/\"\n",
		     $1.tadd, idnum, (int)$5.str, xml, LINE_INFO(@$));
	  EFREE (xml);
	  free_lval (&$1); free_lval (&$3);
	  fprintf (yyout, "@\"/cond_selected\"\n");
      }
    ;

else_group
    : /* empty */
      {
	  NULLIFY ($$);
	  $$.tadd = estrdup ("");
      }
    | "#else" NEWLINE
      {
	  fprintf (yyout, "#else\n");
          fprintf (yyout, "@\"cond_selected type='else' ref='F%d_%d' "
		   LINE_FMT "\"\n", idnum, id_counter, LINE_INFO2(@1,@2));
	  $$.str = (char *)id_counter++; 
      }
      group 
      {
	  NULLIFY ($$);
	  easprintf (&$$.tadd, "@\"else id='F%d_%d' " LINE_FMT "/\"\n",
		     idnum, (int)$3.str, LINE_INFO(@$));
	  fprintf (yyout, "@\"/cond_selected\"\n");
      }
    ;

/* for #define's pp_tokens */
pp_tokens_define_opt
    : /* empty */                { BOTTOM($$, ""); }
    | pp_tokens_define           { MOVE($$, $1);   }
    ;

pp_tokens_define
    : preprocessing_token_define { MOVE($$, $1);   }
    | pp_tokens_define preprocessing_token_define 
      {
	  MOVE2_FMT($$, $1, $2, "%s %s");
      }
    ;

/* '#' and '##' are specially treated. */
preprocessing_token_define
    : identifier                  { MOVE($$, $1); }
    | CONSTANT                    { MOVE($$, $1); }
    | STRING_LITERAL              { MOVE($$, $1); }
    | PUNCTUATOR                  { MOVE($$, $1); }
    | NON_WHITE_SPACE_CHAR        { MOVE($$, $1); }
    | '('                         { BOTTOM($$, yytext); }
    | ')'                         { BOTTOM($$, yytext); }
    | "..."                       { BOTTOM($$, yytext); }
    | ','                         { BOTTOM($$, yytext); }
    | '#'  identifier
      {
	  /* eliminate tracers for #'s operands */
	  easprintf (&$$.str,  "#%s", $2.str);
	  easprintf (&$$.str2, "#%s", $2.str2);
	  easprintf (&$$.tadd, "#%s", $2.str); /* $2.str, not $2.tadd */
	  $$.list = NULL; free_lval (&$2);
      }
    | preprocessing_token_define "##" preprocessing_token_define  
      {
	  /* eliminate tracers for ##'s operands */
	  easprintf (&$$.str,  "%s##%s", $1.str,  $3.str);
	  easprintf (&$$.str2, "%s##%s", $1.str2, $3.str2);
	  easprintf (&$$.tadd, "%s##%s", $1.str,  $3.str); /* str, not tadd */
	  $$.list = NULL; free_lval (&$1); free_lval (&$3);
      }
    ;

identifier_list_opt
    : /* empty */
      {
	  $$.str  = estrdup ("");
	  $$.str2 = estrdup ("");
	  $$.tadd = NULL;
	  $$.list = emalloc (sizeof (struct CDLL));
	  CDLL_init ($$.list);
	  cdll_head = $$.list;
      }
    | identifier_list
      {
	  $$.str  = $1.str;
	  $$.str2 = $1.str2;
	  $$.tadd = NULL;
	  $$.list = $1.list;
	  $1.list = NULL;
	  cdll_head = $$.list;
      }
    ;

identifier_list
    : IDENTIFIER
      {
	  $$.str  = estrdup ($1.str);
	  $$.str2 = estrdup ($1.str2);
	  $$.tadd = NULL;
	  $$.list = emalloc (sizeof (struct CDLL));
	  CDLL_init ($$.list); 
	  CDLL_insert ($$.list, estrdup ($1.str));
	  free_lval (&$1);
      }
    | identifier_list ',' IDENTIFIER
      {
	  easprintf (&$$.str,  "%s, %s", $1.str,  $3.str);
	  easprintf (&$$.str2, "%s, %s", $1.str2, $3.str2);
	  $$.tadd = NULL;
	  $$.list = $1.list;
	  CDLL_insert ($$.list, estrdup ($3.str));
	  free_lval (&$1); free_lval (&$3);
      }
    ;

identifier
    : IDENTIFIER
      {
	  $$.str  = $1.str;
	  if ((define_func_flag && CDLL_query (cdll_head, $1.str) != -1)
	      || !strcmp ($1.str, "__VA_ARGS__")) {
	      /* parameter in function-like macro */
	      $$.str2 = $1.str2;
	      easprintf (&$$.tadd, 
			 "@\"macro_call type='param' name='%s' " LINE_FMT "\""
			 "@\"bare\"%s@\"/bare\" @\"double\"%s@\"/double\""
			 "@\"/macro_call\"",
			 $1.str, LINE_INFO(@1), $1.str,
			 ((!strcmp ($1.str, "__VA_ARGS__"))
			  ? $1.str: $1.str2));
	  } else {
	      $$.str2 = $1.str2;
	      easprintf (&$$.tadd, 
			 "@\"macro_call type='obj' name='%s' " LINE_FMT "\""
			 "@\"bare\"%s@\"/bare\" @\"double\"%s@\"/double\""
			 "@\"/macro_call\"",
			 $1.str, LINE_INFO(@1), $1.str, $1.str2);
	  }
	  EFREE ($1.tadd);
	  $$.list = $1.list;
	  NULLIFY ($1);
      }
    ;

/* = pp_tokens_opt - ('(' + ')' + ',') */
preprocessing_token_minus_paren_comma
    : identifier            { MOVE($$, $1); }
    | CONSTANT              { MOVE($$, $1); }
    | STRING_LITERAL        { MOVE($$, $1); }
    | PUNCTUATOR            { MOVE($$, $1); }
    | NON_WHITE_SPACE_CHAR  { MOVE($$, $1); }
    | "..."                 { BOTTOM($$, yytext); }
    | '#'                   { BOTTOM($$, yytext); }
    | "##"                  { BOTTOM($$, yytext); }
    ;

/* = pp_tokens_opt - ('(' + ')') */
preprocessing_token_minus_paren
    : preprocessing_token_minus_paren_comma  { MOVE($$, $1); }
    | ','                                    { BOTTOM($$, yytext); }
    ;

pp_tokens_opt
    : /* empty */           { BOTTOM($$, ""); }
    | pp_tokens             { MOVE($$, $1); }
    ;

pp_tokens
    : preprocessing_token            { MOVE($$, $1); }
    | pp_tokens preprocessing_token  { MOVE2_FMT($$, $1, $2, "%s %s"); }
    ;

preprocessing_token
    : preprocessing_token_minus_paren            { MOVE($$, $1); }
    | func_macro_call                            { MOVE($$, $1); }
    | '(' { BOTTOM($$, yytext); }
    | ')' { BOTTOM($$, yytext); }
/*
    | '(' pp_tokens_minus_paren_plus_nl_opt ')'  { MOVE1_FMT($$, $2, "(%s)"); }
*/
    ;

/* we don't have to take care of IDENTIFIER2 here, since IDENTIFIER2
   will appear only as #define's next identifier. */
func_macro_call
    : IDENTIFIER '(' argument_list ')'
      {
	  char *xml = str_to_xml ($3.str);
	  easprintf(&$$.str,  "%s(%s)", $1.str,  $3.str);
	  easprintf(&$$.str2, "%s(%s%s%s)", $1.str2, $3.str, 
		    ((!strcmp($3.str, "")) ? "" : ", "), $3.str2);
	  if (!strcmp ($1.str, "_Pragma")) {
	      $$.tadd = estrdup ($$.str);
	  } else if (!strcmp ($1.str, "defined") && is_if_or_elif) {
	      easprintf(&$$.tadd,
			"@\"macro_call type='defined' name='%s' args='%s' "
			LINE_FMT "\""
			"@\"bare\"\n#ifdef %s\n1\n#else\n0\n#endif\n@\"/bare\""
			"@\"double\"%s@\"/double\""
			"@\"/macro_call\"",
			$1.str, xml, LINE_INFO(@$), $3.str, $3.tadd);
	  } else if (args.conservative) {
	      easprintf(&$$.tadd,
			"@\"macro_call type='func' name='%s' args='%s' "
			LINE_FMT "\""
			"@\"bare\"%s(%s)@\"/bare\""
			"@\"double\"%s(%s)@\"/double\""
			"@\"/macro_call\"",
			$1.str, xml, LINE_INFO(@$), $1.str, $3.str,
			$1.str, $3.str2);
	  } else {
	      easprintf(&$$.tadd,
			"@\"macro_call type='func' name='%s' args='%s' "
			LINE_FMT "\""
			"@\"bare\"%s(%s)@\"/bare\""
			"@\"double\"%s(%s%s%s)@\"/double\""
			"@\"/macro_call\"",
			$1.str, xml, LINE_INFO(@$), $1.str, $3.str,
			$1.str2, $3.str,
			((!strcmp($3.str, "")) ? "" : ", "), $3.tadd);
	  }
	  EFREE (xml);
	  $$.list = NULL;
	  id_counter++; free_lval (&$1); free_lval (&$3);
      }
    ;

pp_tokens_minus_paren_comma_plus_nl_opt
    : /* empty */                           { BOTTOM($$, ""); }
    | pp_tokens_minus_paren_comma_plus_nl   { MOVE($$, $1); }
    ;
    
pp_tokens_minus_paren_comma_plus_nl
    : preprocessing_token_minus_paren_comma_plus_nl  { MOVE($$, $1); }
    | pp_tokens_minus_paren_comma_plus_nl
      preprocessing_token_minus_paren_comma_plus_nl
      {
	  MOVE2_FMT($$, $1, $2, "%s %s");
      }
    ;

preprocessing_token_minus_paren_comma_plus_nl
    : preprocessing_token_minus_paren_comma     { MOVE($$, $1); }
    | func_macro_call                           { MOVE($$, $1); }
    | '(' pp_tokens_minus_paren_plus_nl_opt ')' { MOVE1_FMT($$, $2, "(%s)"); }
    | NEWLINE                                   { BOTTOM($$, ""); }
    ;

pp_tokens_minus_paren_plus_nl_opt
    : /* empty */                     { BOTTOM($$, ""); }
    | pp_tokens_minus_paren_plus_nl   { MOVE($$, $1); }
    ;
    
pp_tokens_minus_paren_plus_nl
    : preprocessing_token_minus_paren_plus_nl { MOVE($$, $1); }
    | pp_tokens_minus_paren_plus_nl preprocessing_token_minus_paren_plus_nl
      {
	  MOVE2_FMT($$, $1, $2, "%s %s");
      }
    ;

preprocessing_token_minus_paren_plus_nl
    : preprocessing_token_minus_paren_comma_plus_nl  { MOVE($$, $1); }
    | ','                                            { BOTTOM($$, yytext); }

argument_list
    : pp_tokens_minus_paren_comma_plus_nl_opt
      {
	  char *xml = str_to_xml ($1.str);
	  $$.str  = $1.str;
	  $$.str2 = $1.str2;
	  if (!strcmp ($1.str, "")) {
	      $$.tadd = xml;
	  } else {
	      easprintf (&$$.tadd, "@\"arg exp='%s' " LINE_FMT "\"%s@\"/arg\"",
			 xml, LINE_INFO(@1), $1.tadd);
	      EFREE (xml);
	  }
	  $$.list = NULL; EFREE ($1.tadd);
	  NULLIFY ($1);
      }
    | argument_list ',' pp_tokens_minus_paren_comma_plus_nl_opt
      {
	  char *xml = str_to_xml ($3.str);
	  easprintf (&$$.str,  "%s, %s", $1.str,  $3.str);
	  easprintf (&$$.str2, "%s, %s", $1.str2, $3.str2);
	  if (!strcmp ($1.str, "")) {
	      $$.tadd = xml;
	  } else {
	      easprintf (&$$.tadd, "%s, @\"arg exp='%s' " LINE_FMT
			 "\"%s@\"/arg\"",
			 $1.tadd, xml, LINE_INFO(@3), $3.tadd);
	      EFREE (xml);
	  }
	  $$.list = NULL; free_lval (&$1); free_lval (&$3);
	  NULLIFY ($1); NULLIFY ($3);
      }
    ;
%% /* ==================================================================== */
/* to make sure yyparse () does always fetch a look-ahead token eagerly. */
static void
lex_sync (void)
{
    if (yychar == YYEMPTY) {
        yychar = yylex ();
        /* just in case yylex returns negative to indicate EOF */
        if (yychar < 0) 
            yychar = 0; /* EOF */
    }
}

static void
free_lval (struct lval *lval)
{
    assert (lval != NULL);
    if (lval->str  != NULL) { EFREE (lval->str);  lval->str  = NULL; }
    if (lval->str2 != NULL) { EFREE (lval->str2); lval->str2 = NULL; }
    if (lval->tadd != NULL) { EFREE (lval->tadd); lval->tadd = NULL; }
}

static void
free_lval2 (struct lval *lval)
{
    assert (lval != NULL);
    free_lval (lval);
    if (lval->list != NULL) {
	CDLL_free (lval->list);
	lval->list = NULL;
    }
}

static int
yyerror (char *mesg)
{
    switch (yychar) {
    case INCLUDE: case INCLUDE_NEXT: case DEFINE: case UNDEF:
    case IF:      case IFDEF:        case IFNDEF: case ELIF:   case ELSE:
    case ENDIF:   case LINE:         case ERROR:  case PRAGMA: case EMPTY:
	tadd_unput (yytext);
	tadd_unput ("\n");
	yylloc.last_line--;
	yyclearin;
	break;
    default:
	/* do nothing */
	break;
    }

    return fprintf (stderr, "error(%d: %d.%d-%d.%d): %s\n",
		    LINE_INFO(yylloc), mesg);
}

static void
print_usage (void)
{
    fprintf (stderr, 
	     "usage  : tadd.exe [options] foo.[ch] ...\n"
	     "options:\n"
	     "    --dir         dir      output dir [./TBCppA]\n"
	     "    --db-file     file     db_flie [db_file](relative to dir)\n"
	     "    --comm-suffix suffix   suffix [.comm]\n"
	     "    --double-suffix suffix [_1528552]\n"
             "    --yydebug              enable bison debugging[off]\n"
	     "    --conservative         no embedding tracers to macro args\n"
	     "    --help                 prints this usage\n"
	     );
}

static void
process_options (int argc, char *argv [])
{
    int    i;
    for (i = 1; i < argc; i++) {
	if (argv [i][0] != '-') {                    /* filenames */
	    int j;
	    for (j = 0; i < argc && argv [i][0] != '-'; i++, j++) {
		args.file = erealloc (args.file, sizeof (char *) * (j + 1));
		args.file [j] = emalloc (strlen (argv [i]) + 1);
		strcpy (args.file [j], argv [i]);
	    }
	    i--;
	    args.file_num = j;
	} else if (!strcmp (argv [i], "--dir")) {
	    i++;
	    args.dir = argv [i];
	} else if (!strcmp (argv [i], "--db-file")) {
	    i++;
	    args.db_file = argv [i];
	} else if (!strcmp (argv [i], "--comm-suffix")) {
	    i++;
	    args.comm_suffix = argv [i];
	} else if (!strcmp (argv [i], "--yydebug")) {
	    yydebug = 1;
	} else if (!strcmp (argv [i], "--double-suffix")) {
	    i++;
	    args.double_suffix = argv [i];
	} else if (!strcmp (argv [i], "--conservative")) {
	    args.conservative = 1;
	} else {
	    if (strcmp (argv [i], "--help"))
		fprintf (stderr, "unknown option: %s\n", argv [i]);
	    print_usage ();
	    exit (1);
	}
    }
}

static void
output_xml_begin (char *file)
{
    fprintf (yyout, "@@preamble %s.xml %s%s\n",
	     args.db_file, file, args.comm_suffix);
    fprintf (yyout, 
	     "@\"file id='F%d_%d' file='%s' path='%s' filename_ref='F%d'\"\n",
	     idnum, id_counter++, file, absolute_path, idnum);
}

static void
output_xml_end (void)
{
    fprintf (yyout, "@\"/file\"\n");
}

static void
output_params (struct CDLL *head)
{
    struct CDLL *p;
    for (p = head->next; p != head; p = p->next)
	fprintf (yyout, "@\"param name='%s'/\"\n", p->param_name);
}

static void
output_params2 (struct CDLL *head)
{
    struct CDLL *p;
    for (p = head->next; p != head; p = p->next)
	fprintf (yyout, "@\"param name='%s'\"%s@\"/param\"",
		 p->param_name, p->param_name);
}

static void
output_error_text (int idnum, int first_line, int first_column,
		   int last_line, int last_column)
{
    int c, line = 1, column = 0;
    FILE *fp = fopen (absolute_path, "r");
    while ((c = getc (fp)) != EOF) {
	if ((first_line < line && line < last_line)
	    || (line == first_line && column >= first_column)
	    || (line == last_line  && column < last_column)) {
	    putc (c, yyout);
	}
	if (c == '\n') {
	    line++;
	    column = 0;
	} else {
	    column++;
	}
    }
    fclose (fp);
}

static void
tadd (char *file)
{
    char *path, *comm, *ret;

    ret = realpath (file, absolute_path);
    if (ret == NULL) {
	perror ("realpath"); exit (1);
    }
    pi_add (file, absolute_path);
    idnum = pi_path2idnum (absolute_path);

    yyin = fopen (file, "r");
    if (yyin == NULL) {
	perror ("fopen"); return;
    }

    easprintf (&path, "%s/%s", args.dir, file);
    easprintf (&comm, "%s/%s%s", args.dir, file, args.comm_suffix);

    yyout   = fopen_with_mkdir (path, "w");
    if (yyout == NULL) {
	perror ("fopen"); exit (1);
    }

    commout = fopen_with_mkdir (comm, "w");
    if (commout == NULL) {
	perror ("fopen"); exit (1);
    }

    yylloc.first_line   = yylloc.last_line   = 1;
    yylloc.first_column = yylloc.last_column = 0;
    
    output_xml_begin (file);
    yyparse ();
    output_xml_end   ();

    fclose (yyin); fclose (yyout); fclose (commout);
    EFREE (path); EFREE (comm);
}

/* ==================================================================== */
int
main (int argc, char *argv [])
{
    int i;
    char *db_path, *dir;

#if defined(USE_TEF)
    TEF_is_in_main = 1;
#endif

    dir = getenv ("TBCPPA_DIR");
    if (dir != NULL)
	args.dir = dir;

    process_options (argc, argv);

    easprintf (&db_path, "%s/%s", args.dir, args.db_file);
    pi_init (db_path);
    EFREE (db_path);
    absolute_path = emalloc (path_max + 1); /* this must be after pi_init() */

    for (i = 0; i < args.file_num; i++) {
	tadd (args.file [i]);
    }

    pi_final ();
    EFREE (absolute_path);
    return 0;
}

