<?xml version="1.0"?>
<xsl:stylesheet version="1.0" xmlns:xsl="http://www.w3.org/1999/XSL/Transform">
<!-- $Id: tifexp.xsl,v 1.1 2006/06/08 17:33:49 gondow Exp gondow $ -->
<xsl:output method="text"/>
<xsl:param name="line"/>
<xsl:param name="filename"/>

<xsl:template match="/">
digraph G {
   node  [fontname=Helvetica,fontsize=10,style=filled,shape=box];
   edge  [fontname=Helvetica,fontsize=10];
   graph [fontname=Helvetica,fontsize=10];
   <xsl:apply-templates select="//cond_selected[
	 (id(@filename_ref)/@file = $filename) and
	(@first_line &lt;= $line) and ($line &lt;= @last_line)]/exp"/>
}
</xsl:template>

<xsl:template match="exp">
  <xsl:variable name="html_filename"><xsl:value-of select="id(../@filename_ref)/@file"/>.html#<xsl:value-of select="../@first_line"/></xsl:variable>
  <xsl:value-of select="generate-id(.)"/> [label="#<xsl:value-of select="../@type"/><xsl:text>  </xsl:text><xsl:value-of select='@orig'/>",color=lightblue,URL="./<xsl:value-of select='$html_filename'/>"];
  <xsl:apply-templates/>
</xsl:template>

<xsl:template match="text()">
  <xsl:if test="normalize-space(.)!=''">
    <xsl:value-of select="generate-id(.)"/> [label="<xsl:value-of select='normalize-space(.)'/>",color=black,shape=ellipse,style=solid];
    <!-- -->
    <xsl:choose>
    <xsl:when test="name(..)='exp' or name(..)='macro_body' or name(..)='macro_call' or name(..)='arg'">
      <xsl:value-of select="generate-id(..)"/> -&gt; <xsl:value-of select="generate-id(.)"/>;
    </xsl:when>
    <xsl:when test="name(../..)='macro_call'">
      <xsl:value-of select="generate-id(../..)"/> -&gt; <xsl:value-of select="generate-id(.)"/>;
    </xsl:when>
    <xsl:otherwise>
      <xsl:message>ERROR: wrong text() match: <xsl:value-of select="name(..)"/> <xsl:value-of select="name(../..)"/>
      </xsl:message>
    </xsl:otherwise>
    </xsl:choose>
  </xsl:if>
</xsl:template>

<xsl:template match="macro_call">
  <xsl:choose>
  <xsl:when test="name(..)='exp' or name(..)='macro_body' or name(..)='macro_call' or name(..)='arg'">
    <xsl:value-of select="generate-id(..)"/> -&gt; <xsl:value-of select="generate-id(.)"/>;
  </xsl:when>
  <xsl:when test="name(../..)='macro_call'">
    <xsl:value-of select="generate-id(../..)"/> -&gt; <xsl:value-of select="generate-id(.)"/>;
  </xsl:when>
  </xsl:choose>

  <xsl:variable name="html_filename"><xsl:value-of select="id(@filename_ref)/@file"/>.html#<xsl:value-of select="@first_line"/></xsl:variable>

  <xsl:choose>
  <xsl:when test="@type='func'">
    <xsl:value-of select="generate-id(.)"/> [label="[func macro]\n<xsl:value-of select='@name'/>(<xsl:value-of select='normalize-space(@args)'/>)",URL="./<xsl:value-of select='$html_filename'/>",color=lightpink,peripheries=2];
  </xsl:when>
  <xsl:when test="@type='obj'">
    <xsl:value-of select="generate-id(.)"/> [label="[obj macro]\n<xsl:value-of select='@name'/>",URL="./<xsl:value-of select='$html_filename'/>",color=lightpink];
  </xsl:when>
  <xsl:when test="@type='param'">
    <xsl:value-of select="generate-id(.)"/> [label="[param]\n<xsl:value-of select='@name'/>",URL="./<xsl:value-of select='$html_filename'/>",color=orange];
  </xsl:when>
  <xsl:when test="@type='defined'">
    <xsl:value-of select="generate-id(.)"/> [label="[defined]\ndefined(<xsl:value-of select='@args'/>) = <xsl:value-of select='normalize-space(bare)'/>",URL="./<xsl:value-of select='$html_filename'/>",color=red];
  </xsl:when>
  <xsl:otherwise>
    <xsl:message>
      ERROR: unknown macro_call type: <xsl:value-of select="@type"/>
    </xsl:message>
  </xsl:otherwise>
  </xsl:choose>
  <xsl:apply-templates select="double"/>
</xsl:template>

<xsl:template match="macro_body">
  <xsl:choose>
  <xsl:when test="name(..)='exp' or name(..)='macro_body' or name(..)='macro_call' or name(..)='arg'">
      <xsl:value-of select="generate-id(..)"/> -&gt; <xsl:value-of select="generate-id(.)"/>;
  </xsl:when>
  <xsl:when test="name(../..)='macro_call'">
      <xsl:value-of select="generate-id(../..)"/> -&gt; <xsl:value-of select="generate-id(.)"/>;
  </xsl:when>
  </xsl:choose>

  <xsl:variable name="macro_def" select="id(@ref)"/>
  <xsl:variable name="html_filename"><xsl:value-of select="id($macro_def/@filename_ref)/@file"/>.html#<xsl:value-of select="$macro_def/@first_line"/></xsl:variable>

  <xsl:variable name="sig">
    <xsl:choose>
      <xsl:when test="$macro_def/@type='func'">
        <xsl:value-of select='$macro_def/@name'/>(<xsl:value-of select='normalize-space($macro_def/@args)'/>)</xsl:when>
      <xsl:when test="$macro_def/@type='obj'">
        <xsl:value-of select='$macro_def/@name'/></xsl:when>
    </xsl:choose>
  </xsl:variable>
  <xsl:value-of select="generate-id(.)"/> [label="#define <xsl:value-of select='$sig'/> <xsl:text> </xsl:text><xsl:value-of select='normalize-space($macro_def/@macro_body)'/>",color=lightblue,URL="./<xsl:value-of select='$html_filename'/>"];
  <xsl:apply-templates/>
</xsl:template>

<xsl:template match="arg">
  <xsl:choose>
  <xsl:when test="name(..)='exp' or name(..)='macro_body' or name(..)='macro_call' or name(..)='arg'">
      <xsl:value-of select="generate-id(..)"/> -&gt; <xsl:value-of select="generate-id(.)"/>;
  </xsl:when>
  <xsl:when test="name(../..)='macro_call'">
      <xsl:value-of select="generate-id(../..)"/> -&gt; <xsl:value-of select="generate-id(.)"/>;
  </xsl:when>
  </xsl:choose>

  <xsl:variable name="html_filename"><xsl:value-of select="id(@filename_ref)/@file"/>.html#<xsl:value-of select="@first_line"/></xsl:variable>

  <xsl:value-of select="generate-id(.)"/> [label="[arg]\n<xsl:value-of select='@exp'/>",URL="./<xsl:value-of select='$html_filename'/>",color=greenyellow];

  <xsl:apply-templates/>
</xsl:template>

<xsl:template match="double">
  <xsl:apply-templates/>
</xsl:template>

<xsl:template match="param">
</xsl:template>
</xsl:stylesheet>

<!--
lightblue, lightpink, greenyellow
-->

